% LICENSE AGREEMENT
% - - - - - - - - -
% Copyright (c) 2019, Stilianos Louca
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions are
% met:
%
%    * Redistributions of source code must retain the above copyright
%      notice, this list of conditions and the following disclaimer.
%    * Redistributions in binary form must reproduce the above copyright
%      notice, this list of conditions and the following disclaimer in
%      the documentation and/or other materials provided with the distribution
%    * Neither the name of the original author (Stilianos Louca), nor the names
%      of any other contributors may be used to endorse or promote products derived
%      from this software without specific prior written permission.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
% AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
% ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
% LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
% SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
% INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
% CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
% ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%
%
% DESCRIPTION
% - - - - - -
% This MATLAB code performs Spatial Metabolic Flux (SMF) analysis.
%
% Given a set of metabolites & reactions on a finite 1D interval, with Dirichlet and/or Robin boundary conditions, determine the hotspots and rates of reactions that "best" satisfy the following constraints:
%   1. At any location, no metabolite has negative concentration.
%   2. At any location, and for any reaction, at least one substrate metabolite must be zero (and thus limiting the reaction rate).
% The hotspots and amplitudes of reactions are determined by iteratively minimizing a stress function quantifying the deviation from the above constraints.
% Metabolite transport is assumed to be purely diffusive and/or advective (the diffusivity and advection speed may differ for each metabolite and each location).
% Boundary conditions for metabolites at teh top & bottom can be either Dirichlet (fixed concentration) and/or Robin (fixed flux).
% The underlying reaction-transport model is a steady-state model.
%
% Main inputs:
%	The domain's grid
%	Boundary conditions for each metabolite (Dirichlet and/or Robin)
%   The diffusivity of each metabolite
%	Stoichiometric matrix of all reactions
%	Optional concentration data (scattered or on depth-grid) for model-data comparison during fitting
% Main outputs:
%	The locations of NH hotspots
%	The rate of each reaction at each hotspot
%	The corresponding metabolite concentration profiles
%
% Canonical units assumed for each variable:
%	Depth (location): meters
%	Time: days
%	Diffusivity: m^2/day
%	Advection speed: m/day
%	Dilution area (lateral basin area): arbitrary, for example km^2
%	Metabolite concentrations: mol/L
%	Reaction rates at each hotspot: kmol/(m^2 * day) (normalized w.r.t. the lateral area at each hotspot)
%	Ghost source/sink rates at each hotspot: kmol/(m^2 * day) (normalized w.r.t. the lateral area at each hotspot)
%	Metabolic fluxes at each hotspot: kmol/(m^2 * day) (normalized w.r.t. the lateral area at each hotspot)
%
% For a background on SMF see the following paper:
%	Louca, Stilianos et al. (in review as of 2018). Shortcutting microbial kinetics in biogeochemical models.
%
% Tested on MATLAB R2015b.
%
% - - - - - - - - -
% Code name and version: SMF 2019.02.12
% Tested on MATLAB R2015b.
% - - - - - - - - -
%
function [	hotspots,...			% NH sized vector listing the predicted locations (depths) of hotspots.
			reaction_rates,...		% NR x NH matrix listing predicted reaction rates at each hotspot. Reaction rates are lateral-area-specific, w.r.t. the area at each hotspot depth. If you want to integrate rates over depth, you will need to account for a variable dilution_area if applicable.
			ghost_source_rates,...	% NGsources x NH matrix listing predicted rates of ghost sources. Ghost rates are lateral-area-specific, w.r.t. the area at each hotspot depth.
			ghost_sink_rates,...	% NGsinks x NH matrix listing predicted rates of ghost sinks. Ghost rates are lateral-area-specific, w.r.t. the area at each hotspot depth.
			concentrations,...		% NM x NZ matrix listing the predicted metabolite concentrations (accounting for boundary conditions, reaction rates and any explicit production rates).
			concentrations_BC_part,...					% NM x NZ matrix listing metabolite concentrations solely due to boundary conditions, omitting any internal sources/sinks
			concentrations_added_production_part,... 	% NM x NZ matrix listing metabolite concentrations solely due to added_production_rates, assuming zero boundary conditions
			concentrations_reaction_part,...			% NM x NZ matrix listing metabolite concentrations solely due to reaction rates, assuming zero boundary conditions
			metabolite_net_source_rates,...		% NM x NH matrix listing the net production rate of each metabolite at each hotspot (due to reaction rates and ghost source/sink rates). Metabolite flux rates are lateral-area-specific, w.r.t. the area at each hotspot depth.
			metabolite_boundary_influxes,...	% NM x 2 matrix listing the net vertical flux rate of each metabolite into the system through the top and bottom boundary. Hence, metabolite_boundary_influxes(m,1) is the net influx rate of the m-th metabolite through the top boundary, metabolite_boundary_influxes(m,2) is the net influx rate of the m-th metabolite through the bottom boundary. Boundary influx rates are lateral-area-specific, w.r.t. the area at each boundary.
			stress,...							% scalar, specifying the minimum stress achieved during fitting, and corresponding to the returned solution
			R2s_grid_data,...					% 1D vector of size NM, listing R2s (fractions of explained variance) of the supplied concentration_grid_data{}. For metabolites with missing concentration_grid_data, NaN is returned.
			R2s_scatter_data,...				% 1D vector of size NM, listing R2s (fractions of explained variance) of the supplied concentration_scatter_data{}. For metabolites with missing concentration_scatter_data, NaN is returned.
			hotspots_all_trials,...				% 2D matrix of size Ntrials x NH, listing fitted locations of hotspots for each fitting trial. Only relevant if all_trial_results==true.
			reaction_rates_all_trials,...		% 3D matrix of size Ntrials x NR x NH, listing fitted reaction rates for each fitting trial. Only relevant if all_trial_results==true.
			stress_all_trials,...				% numeric vector of size Ntrials, listing the minimized stress for each converged fitting trial. Only relevant if all_trial_results==true.
			Niterations_all_trials,...			% numeric vector of size Ntrials, listing the number of iterations for each fitting trial. Only relevant if all_trial_results==true.
			Nevaluations_all_trials,...			% numeric vector of size Ntrials, listing the number of evaluations of the stress function for each fitting trial. Only relevant if all_trial_results==true.
			R2s_grid_data_all_trials,...		% 2D matrix of size Ntrials x NM, listing R2s (fractions of explained variance) of the supplied concentration_grid_data{} for each fitting trial and each metabolite. For metabolites with missing concentration_grid_data, NaN is returned.
			R2s_scatter_data_all_trials,...		% 2D matrix of size Ntrials x NM, listing R2s (fractions of explained variance) of the supplied concentration_scatter_data{} for each fitting trial and each metabolite. For metabolites with missing concentration_scatter_data, NaN is returned.
			spurious_hotspots,...				% 1D vector of size <=NH, listing hotspot indices identified as likely spurious due to their convergence to the permissible boundaries. Note that if omit_spurious_hotspots==true, the reaction rates in these hotspots will be set to zero.
			negligible_hotspots,...				% 1D vector of size <=NH, listing hotspot indices identified as likely negligible due to generally low metabolic fluxes (based on the given relative threshold negligible_rel_flux_threshold).
			exit_flag,...						% exit flag from fmincon algorithm, for the best fitting trial. A positive number indicates success. Details can be found here: https://www.mathworks.com/help/optim/ug/fmincon.html#busog7r-exitflag
			Niterations,...						% number of iterations of fmincon algorithm, for the best fitting trial.
			Nfailed_trials,...					% number of fitting trials that failed (i.e. had an exit flag <=0). This will be a number between 0 and Ntrials.
			errormsg] = fitSMFmodel(	gridZ,...					% 1D numeric vector of size NZ, defining the domain grid, with z values in ascending order, including the domain's boundaries [A,B]
										NM, ...						% integer, number of metabolites
										NR, ...						% integer, number of reactions
										NH,... 						% non-negative integer, fixed number of hotspots to include in model. The locations of hotspots, and the reaction rates at each hotspot, are fitted parameters.
										metabolite_names,...		% cell array of size NM, specifying metabolite names. Mostly used for progress reports and error messages
										reaction_names,...			% cell array of size NR, specifying reaction names. Mostly used for progress reports and error messages
										BCtypesA, ...				% boundary condition type for each metabolite at A: 0 for Dirichlet (fixed value), 1 for Robin (fixed flux)
										BCtypesB, ...				% boundary condition type for each metabolite at B: 0 for Dirichlet (fixed value), 1 for Robin (fixed flux)
										BCvaluesA, ...				% boundary values of metabolites at A, interpreted according to BCtypesA. e.g. Dirichlet: mol/L;  Robin (inward flux): kmol/m^2/day
										BCvaluesB, ...				% boundary values of metabolites at B, interpreted according to BCtypesB. e.g. Dirichlet: mol/L;  Robin (inward flux): kmol/m^2/day
										diffusivities, ... 			% 1D cell array of size 1 or NM, defining the diffusivities of metabolites (functional defined on [A,B]). If of size 1, then all metabolites are assumed to have the same diffusivity. If of size 0 (empty cell array), then diffusivity is assumed to be zero for all metabolites. Can also contain scalars instead of function handles, indicating a spatially-constant diffusivity.
										advections, ... 			% 1D cell array of size 1 or NM, defining the advection velocities of metabolites (functional defined on [A,B]). If of size 1, then all metabolites are assumed to have the same advection velocity. If of size 0 (empty cell array), then advection is assumed to be zero for all metabolites. Can also contain scalars instead of function handles, indicating a spatially-constant advection.
										dilution_area, ... 			% (function handle or a single number) lateral (cross-sectional) area of the modeled basin, to account for potential geometric dilution effects. If you're modeling a vertical water column in a restricted basin, and the basin's lateral area changes over depth, you need to account for that variability. The units of this function can be arbitrary (e.g., m^2 or km^2), since the only thing that matters is the ratio between any two depths. To ignore this effect, set this function to 1.
										grid_temperatures,...		% 1D vector of size NZ, listing temperatures on gridZ. Only relevant if penalize_inefficiencies_by=='reaction_DeltaG'.
										use_ficks_law,...			% (boolean) whether to use Fick's law to model diffusion. If true, the diffusive flux is D*p_xC, otherwise it is p_x(D*C).										
										reaction_stoichiometry, ...			% stoichiometric matrix for reactions. 2D array of size NM x NR.
										reaction_DeltaG0,...				% standard Gibbs free energies of reactions. 1D vector of size NR. Only relevant if penalize_inefficiencies_by=='reaction_DeltaG'.
										concentration_grid_data,...			% optional 1D cell array of size NM or 0, listing optional data (raw or interpolated) of metabolite concentrations on gridZ, to be considered in model fitting. Each entry concentration_grid_data{m} is either an empty vector, or a 1D vector of size NZ, listing concentrations for metabolite m at depths gridZ.
										concentration_scatter_data,...		% optional 1D cell array of size NM or 0, listing optional scattered data of metabolite concentrations, to be considered in model fitting. Each entry concentration_scatter_data{m} is either NaN, or an empty vector, or a 2D matrix of size NDm x 2, listing locations (column 1) and values (column 2) of NDm data points for metabolite m.
										added_production_rates,...			% optional 1D cell array of size NM or 0, listing explicitly added net source rates for each metabolite at each grid point (as scalar functionals, negative values indicating net consumption). These net source rates are additional to the reaction-driven source rates at the fitted reaction hot spots, and can be used to inform the model of additional sources/sinks of metabolites not captured by the SMF framework. This array can also be empty, or contain NaN for a particular metabolite, meaning no explicit source rates are added.
										ghost_sources, ... 					% optional 1D integer vector, listing metabolite indices for which to include pure production (dummy) reactions (e.g., modeling unknown metabolite sources). Can be empty ([]).
										ghost_sinks, ... 					% optional 1D integer vector, listing metabolite indices for which to include pure consumption (dummy) reactions (e.g., modeling unknown metabolite sinks). Can be empty ([]).
										metabolite_weights, ...				% either an empty vector, or a vector of size 1, or a vector of size NM listing relative error metabolite_weights for each metabolite. If empty, metabolite_weights are chosen automatically. If of length 1, all metabolites have equal weight. If of length NM, then NaN entries will be set to the non-NaN average (e.g. [1 NaN 2] becomes [1 1.5 2])..
										positivity_scales,...				% reference scales for metabolites for penalizing positive concentrations due to inefficient resource use. Typically comparable to half-saturation constants or boundary concentrations. Either an empty vector, or a vector of size 1, or a vector of size NM. May only contain strictly positive numbers or NaNs (these will be set to the mean concentration of the reactionless solutions).
										negativity_scales,...				% reference scales for metabolites for penalizing negative concentrations. Typically comparable to the boundary concentrations. Either an empty vector, or a vector of size 1, or a vector of size NM. May only contain strictly positive numbers or NaNs (these will be set to the mean concentration of the reactionless solutions).
										penalize_inefficiencies_by,...		% (string) how to penalize inefficient substrate use, i.e. when limiting substrates are not zero. Can be 'none', 'reaction_DeltaG', 'substrate_geomean' or 'min_substrate_concentrations'.
										negative_concentration_penalty,...	% factor by which negative metabolite concentrations should be pelanized. If Inf, then the requirement of non-negative concentrations is hardwired as a hard constraint into the optimizer.
										negative_flux_penalty,...			% factor by which metabolite sinks should be pelanized when they occur in regions of negative metabolite concentrations.
										include_grid_data,...				% include deviations from concentration grid-data in stress function (minimized during fitting)
										include_scatter_data,...			% include deviations from concentration scatter-data in stress function (minimized during fitting)
										start_hotspots, ...					% optional 1D numeric vector of size NH, specifying starting (initial guess) values for hotspot locations. If empty, this is chosen heuristically. May include NaNs (these hotspots will always start at random values).
										start_reaction_rates, ...			% optional 2D numeric matrix of size NR x NH, specifying starting (initial guess) values for reaction rates. If empty, this is chosen heuristically. May include NaNs (these rates will always start at random values).
										always_use_start_hotspots,...		% optional 1D boolean vector of size NH, specifying if any of the provided start_hotspots should be used as start value for all fitting trials. For example, if [true false], then the 1st start hotspot is used in all trials, whereas the 2nd start hotspot is only used in the first trial. Can also be a single boolean, applying to all hotspots.
										always_use_start_reaction_rates,...	% optional 2D boolean matrix of size NR x NH, specifying which of the provided start_reaction_rates should be used as start value for all fitting trials. For example, if [true; false], then start_reaction_rates(1,1) is used in all trials, whereas start_reaction_rates(2,1) is only used in the first trial. Can also be a single boolean, applying to all reaction rates.
										fixed_hotspots,...					% optional 1D numeric vector of size NH, specifying fixed locations for a subset of hotspots. A non-NaN value means the hotspot has a fixed (non-fitted) location; a NaN value means the hotspot's location will be fitted. Can also be empty ([]), in which case all hotspot locations will be fitted.
										fixed_reaction_rates,...			% optional 2D numeric array of size NR x NH, specifying fixed locations for a subset of reaction rates. A non-NaN value means the reaction has a fixed (non-fitted) rate; a NaN value means the reaction rate will be fitted. Can also be empty ([]), in which case all reaction rates will be fitted.
										allowed_hotspot_range,...			% optional 1D numeric vector of size 2, listing min & max allowed hotspot depths. If empty ([]), hotspots are in principle allowed anywhere in the modeled depth range. May include NaN or +Inf or -Inf.
										optim_method,...					% string, specifying the optimization method to use. Options are 'fmincon' or 'patternsearch'
										optim_max_iter,...					% positive integer, maximum number of iterations during fitting
										optim_max_evals,...					% positive integer, maximum number of function evaluations during fitting
										objective_tolerance,...				% relative tolerance of objective function during fitting ('TolFun'/stress_scales). A typical value would be 1e-6.
										parameter_tolerance,...				% relative tolerance of fitted parameters ('TolX'/param_scales). A typical value would be 1e-6.
										concentration_epsilon,...			% threshold below which metabolite concentrations are essentially zero. For numerical rounding purposes only, not used to weigh the stress function.
										max_random_rate_factor,...			% (numeric) maximum factor by which to randomly increase non-fixed start rates for each fitting trial, compared to their default start value. If start values are provided, random variation is applied w.r.t those start values; a higher max_random_rate_factor would reflect a higher uncertainty in the start values. A typical value would be between 10-1000.
										Ntrials, ...						% (integer) number of random fitting trials, i.e. starting at random params (for any given number of hotspots)
										Nrounds,...							% (integer) number of rounds of a whole batch of fitting trials, e.g. if the best fit from the last round yielded spurious hotspots. Hence, the total number of fitting trials may end up being as high as Nrounds * Ntrials
										Nstart_attempts,...					% (integer) number of times that random start params should be generated until a valid starting point is found (i.e. returning a valid stress value and if possible satisfying constraints), for each fitting trial. If this number is exceeded without finding a valid starting point (i.e. all starting points returned an invalid stress value), the particular fitting trial is skipped. Starting points violating constraints, but still returning a valid stress value, are accepted if needed. Typical values are 1-100.
										parallel,...						% (boolean) whether to run fitting trials in parallel using multiple threads
										omit_spurious_hotspots,...			% (boolean) omit spurious hotspots from the final output
										negligible_rel_flux_threshold,...	% (numeric) relative metabolite flux threshold below which a hotspot is identified as "negligible" (and is thus omitted from the final solution). Set this to negative to disable this filter.
										R2s_for_all_trials,...				% (boolean) return R2s w.r.t. data for all fitting trials. If false, R2s_grid_data_all_trials and R2s_scatter_data_all_trials will be NaN.
										check_nans_in_input,...				% (boolean) check if input data contains NaNs where it shouldn't. If check_nans_in_input==false, input data is blindly assumed to not contain NaNs
										verbose, ...						% (boolean) if true, progress messages will be displayed
										verbose_prefix)						% (string) prefix for progress messages. E.g. "  "
	NZ 				= length(gridZ);
	zA 				= gridZ(1);
	zB 				= gridZ(end);
	NGSources 		= length(ghost_sources);
	NGSinks   		= length(ghost_sinks);
	Nrounds 		= max(1,Nrounds);
	Nstart_attempts = max(1,Nstart_attempts);
	
	% hardwired technical options
	NEGATIVE_CONCENTRATION_PENALTY_POWER 	= 2;
	NEGATIVE_FLUX_PENALTY_POWER 			= 2;
	INEFFICIENCY_PENALTY_POWER 				= 2;
	INTERPOLATE_LOCAL_GREENS_FUNCTIONS		= true;	% whether to linearly interpolate Green's functions between adjacent grid points. If false, a more complex algorithm is used to determine Green's functions localized between grid points. After some testing, it seems that this does not noticeably affect the outcome.
	BOUNDARY_EPSILON_FOR_SPURIOUS_HOTSPOTS	= mean(diff(gridZ)); % small depth range, counting from the boundaries, within which to identify spurious hotspots (i.e. hotspots that converged to the boundary, potentially because they confer no improvement to the stress function)
	
	% re-shape some input arrays to ensure proper format
	start_hotspots 			 = reshape(start_hotspots, [length(start_hotspots) 1]);
	always_use_start_hotspots = reshape(always_use_start_hotspots, [length(always_use_start_hotspots) 1]);
	fixed_hotspots 			 = reshape(fixed_hotspots, [length(fixed_hotspots) 1]);
	ghost_sources 			 = reshape(ghost_sources, [1 length(ghost_sources)]);
	ghost_sinks 			 = reshape(ghost_sinks, [1 length(ghost_sinks)]);
	gridZ					 = reshape(gridZ, [NZ 1]);
	
	% default return values
	hotspots 						= NaN;
	reaction_rates 					= NaN;
	ghost_source_rates				= NaN;
	ghost_sink_rates				= NaN;
	concentrations 					= NaN;
	concentrations_BC_part			= NaN;
	concentrations_added_production_part = NaN;
	concentrations_reaction_part	= NaN;
	metabolite_net_source_rates		= NaN;
	metabolite_boundary_influxes	= NaN;
	stress 							= NaN;
	R2s_grid_data					= NaN;
	R2s_scatter_data				= NaN;
	hotspots_all_trials				= NaN;
	reaction_rates_all_trials		= NaN;
	stress_all_trials				= NaN;
	Niterations_all_trials			= NaN;
	Nevaluations_all_trials			= NaN;
	R2s_grid_data_all_trials		= NaN;
	R2s_scatter_data_all_trials		= NaN;
	spurious_hotspots				= [];
	negligible_hotspots				= [];
	exit_flag						= NaN;
	Niterations						= NaN;
	Nfailed_trials					= 0;
	errormsg 						= '';
	
	% define hotspot range limits
	if(length(allowed_hotspot_range)==0)
		allowed_hotspot_range = [zA, zB];
	elseif(length(allowed_hotspot_range)==2)
		if(isnan(allowed_hotspot_range(1))) allowed_hotspot_range(1) = zA; end
		if(isnan(allowed_hotspot_range(2))) allowed_hotspot_range(2) = zB; end
		allowed_hotspot_range(1) = max(zA,min(zB,allowed_hotspot_range(1)));
		allowed_hotspot_range(2) = max(zA,min(zB,allowed_hotspot_range(2)));
	else
		errormsg = sprintf('Invalid size for allowed_hotspot_range (%d); must be 0 or 2',length(allowed_hotspot_range));
		return;	
	end
		
	% basic input error checking
	if(not(strcmp(optim_method,'fmincon')) & not(strcmp(optim_method,'patternsearch')))
		errormsg = sprintf('Invalid optimization method "%s"',optim_method);
		return;
	end
	if((length(start_hotspots)>0) & (length(start_hotspots)~=NH))
		errormsg = sprintf('Number of provided start hotspots (%d) differs from the number of modeled hotspots (%d)',length(start_hotspots),NH);
		return;
	end
	if((length(start_hotspots)>0) & (any(start_hotspots<allowed_hotspot_range(1)) | any(start_hotspots>allowed_hotspot_range(2))))
		errormsg = sprintf('Some of start_hotspots are outside of the allowed depth range (%g-%g)',allowed_hotspot_range(1),allowed_hotspot_range(2));
		return;
	end
	if((length(start_reaction_rates)>0) & ((size(start_reaction_rates,2)~=NH) | (size(start_reaction_rates,1)~=NR)))
		errormsg = sprintf('Shape of provided start reaction rates (%d x %d) differs from expectation (%d x %d)',size(start_reaction_rates,1),size(start_reaction_rates,2),NR,NH);
		return;
	end
	if(not(use_ficks_law))
		errormsg = 'Only Fickian diffusion is supported. Non-Fickkian diffusion has not been implemented yet';
		return;
	end
	if((length(fixed_hotspots)>0) & (length(fixed_hotspots)~=NH))
		errormsg = sprintf('Length of fixed_hotspots (%d) differs from the number of modeled hotspots (%d)',length(fixed_hotspots),NH);
		return;
	end
	if((length(fixed_hotspots)>0) & (any(fixed_hotspots<allowed_hotspot_range(1)) | any(fixed_hotspots>allowed_hotspot_range(2))))
		errormsg = sprintf('Some of fixed_hotspots are outside of the allowed depth range (%g-%g)',allowed_hotspot_range(1),allowed_hotspot_range(2));
		return;
	end
	if((length(fixed_reaction_rates)>0) & ((size(fixed_reaction_rates,2)~=NH) | (size(fixed_reaction_rates,1)~=NR)))
		errormsg = sprintf('Shape of fixed_reaction_rates (%d x %d) differs from expectation (%d x %d)',size(fixed_reaction_rates,1),size(fixed_reaction_rates,2),NR,NH);
		return;
	end
	if(length(always_use_start_hotspots)>0)
		if(length(always_use_start_hotspots)==1)
			always_use_start_hotspots = repmat(always_use_start_hotspots,NH,1);
		elseif(length(always_use_start_hotspots)~=NH)
			errormsg = sprintf('Length of always_use_start_hotspots (%d) differs from the number of modeled hotspots (%d)',length(always_use_start_hotspots),NH);
			return;
		end
	end
	if(length(always_use_start_reaction_rates)>0)
		if(length(always_use_start_reaction_rates)==1)
			always_use_start_reaction_rates = repmat(always_use_start_reaction_rates,NR,NH);
		elseif((size(always_use_start_reaction_rates,2)~=NH) | (size(always_use_start_reaction_rates,1)~=NR))
			errormsg = sprintf('Shape of always_use_start_reaction_rates (%d x %d) differs from expectation (%d x %d)',size(always_use_start_reaction_rates,1),size(always_use_start_reaction_rates,2),NR,NH);
			return;
		end
	end

	% sort start_hotspots and fixed_hotspots to be in ascending order, or check that they are already in ascending order
	if(length(start_hotspots)>0)
		if(length(fixed_hotspots)>0)
			% make sure start_hotspots and fixed_hotspots are consistent in terms of order
			composite_hotspots = start_hotspots; composite_hotspots(not(isnan(fixed_hotspots))) = fixed_hotspots(not(isnan(fixed_hotspots)));
			if(not(all(isnan(composite_hotspots))))
				if(any(diff(composite_hotspots(not(isnan(composite_hotspots))))<0))
					errormsg = sprintf('Provided start hotspots and fixed hotspots are inconsistent; their combination must yield a list of hotspot locations in ascending order');
					return;
				end
			end
		end
		if(any(isnan(start_hotspots)))
			if(not(all(isnan(start_hotspots))))
				% some hotspots are NaN and others ar not NaN, so order of non-NaNs must be ascending already
				if(any(diff(start_hotspots(not(isnan(start_hotspots))))<0))
					errormsg = sprintf('Provided start hotspots must be in ascending order, when NaNs are included');
					return;
				end
			end
		else
			[start_hotspots, order] = sort(start_hotspots);
			if(length(start_reaction_rates)>0)
				start_reaction_rates = start_reaction_rates(:,order);
			end
			if(length(always_use_start_hotspots)>0)
				always_use_start_hotspots = always_use_start_hotspots(order);
			end
			if(length(always_use_start_reaction_rates)>0)
				always_use_start_reaction_rates = always_use_start_reaction_rates(:,order);
			end
		end
	end
	if(length(fixed_hotspots)>0)
		if(any(isnan(fixed_hotspots)))
			if(not(all(isnan(fixed_hotspots))))
				% some hotspots are NaN and others ar not NaN, so order of non-NaNs must be ascending already
				if(any(diff(fixed_hotspots(not(isnan(fixed_hotspots))))<0))
					errormsg = sprintf('Provided fixed hotspots must be in ascending order, when NaNs are included');
					return;
				end
			end
		else
			[fixed_hotspots, order] = sort(fixed_hotspots);
			if(length(fixed_reaction_rates)>0)
				fixed_reaction_rates = fixed_reaction_rates(:,order);
			end
		end
	end

		
	% Extend set of reactions & reaction_stoichiometryiometric matrix to include ghost sources & sinks
	% For example, some metabolites may be allowed to be supplied (consumed) "physically", not just microbiologically
	% This is the same as having additional dummy forward-reactions (with rate P_m) that produce (consume) these metabolites out of nothing
	% hence: S*H + PS + PD = F, where PS_m>=0 if m can be physically supplied (and 0 otherwise), and PD_m>=0 if m can be physically repleted (and 0 otherwise)
	% this can be re-written as Sextended*Hextended = F, where each physically repleted or depleted metabolite gets an additional column in Sextended
	NRe = NR+NGSources+NGSinks;
	start_rates = start_reaction_rates;
	fixed_rates = fixed_reaction_rates;
	always_use_start_rates = always_use_start_reaction_rates;
	if(NRe>NR)
		stoichiometry = [reaction_stoichiometry zeros(NM,NGSources+NGSinks)];
		for g=1:NGSources
			stoichiometry(ghost_sources(g),NR+g) = 1;
		end
		for g=1:NGSinks
			stoichiometry(ghost_sinks(g),NR+NGSources+g) = -1;
		end
		if(length(start_reaction_rates)>0)
			start_rates = [start_reaction_rates; zeros(NGSources+NGSinks,NH)];
		end
		if(length(always_use_start_reaction_rates)>0)
			always_use_start_rates = [always_use_start_reaction_rates; zeros(NGSources+NGSinks,NH)];
		end
		if(length(fixed_reaction_rates)>0)
			fixed_rates = [fixed_reaction_rates; repmat(NaN,NGSources+NGSinks,NH)];
		end
	else
		stoichiometry = reaction_stoichiometry;
	end
	
	
	% pre-process metabolite concentration scatter-data (if available) for faster interpolation
	% in particular, determine which data points are within the considered domain, and find the nearest indices on the grid corresponding to the data points
	NMscatter_data = 0;
	if(length(concentration_scatter_data)>0)
		if(verbose) fprintf('%sPre-processing concentration scatterdata..\n', verbose_prefix); end
		parsed_concentration_scatter_data = cell(NM,1);
		for m=1:NM
			if(size(concentration_scatter_data{m},2)>1)
				NMscatter_data = NMscatter_data + 1;
				xy = concentration_scatter_data{m};
				valids = find((~isnan(xy(:,1))) & (xy(:,1)>=zA) & (xy(:,1)<=zB) & (~isnan(xy(:,2)))); % find data points within domain
				if(length(valids)>0)
					rights = zeros(length(valids),1); % rights(v) will be the index of the grid index nearest (to the right) to the v-th valid data point
					for v=1:length(valids)
						rights(v) = find(gridZ >= xy(valids(v),1), 1, 'first');
					end
					parsed_concentration_scatter_data{m} = [rights xy(valids,:)];
				end
			end
		end
	end
	include_scatter_data = ((NMscatter_data>0) & include_scatter_data);
	
	% pre-process concentration grid-data
	NMgrid_data = 0;
	if(length(concentration_grid_data)>0)
		for m=1:NM
			if(length(concentration_grid_data{m})>0)
				if(all(isnan(concentration_grid_data{m})))
					concentration_grid_data{m} = []; % completely omit this concentration_grid_data, since all NaN
				else
					NMgrid_data = NMgrid_data + 1;
					concentration_grid_data{m} = reshape(concentration_grid_data{m},[1 NZ]); % make sure array is row vector
				end
			end
		end
	end
	include_grid_data = ((NMgrid_data>0) & include_grid_data);

	% get fundamental solution (Green's function) for each metabolite, i.e. steady-state profile that would result if a point-source was located on any of the grid points, with zero boundary conditions (Dirichlet and/or Robin)
	% this will produce an NM x NZ x NZ matrix
	% greens_function[m,s,t] will be the value of the metabolite m at target point t if there was a mass-1 point source at point s, with zero boundary conditions
	% also pre-calculate particular solutions (source-free profiles accounting for boundary conditions and expliti source terms)
	if(verbose) fprintf('%sCalculating Green''s functions for %d metabolites..\n', verbose_prefix, NM); end
	greens = zeros(NM,NZ,NZ);
	concentrations_BC_part = zeros(NM,NZ);
	concentrations_added_production_part = zeros(NM,NZ);
	green_E  			= zeros(NM,NZ);
	green_eV 			= zeros(NM,NZ);
	green_Er 			= zeros(NM,NZ);
	green_W	 			= zeros(NM,NZ);
	green_adv			= zeros(NM,NZ);
	grid_diffusivities 	= zeros(NM,NZ);
	grid_advections		= zeros(NM,NZ);
	diffusive			= zeros(NM,1);
	if(isa(dilution_area,'function_handle'))
		W = dilution_area(zA)./dilution_area(gridZ);
		if(numel(W)==1) W = repmat(W, size(gridZ)); end
	else
		W = ones(size(gridZ));
	end
	precalculations_OK = true;
	for m=1:NM
		% get diffusivity for this metabolite
		if(length(diffusivities)==1)
			diffusivity = diffusivities{1};
		elseif(length(diffusivities)==0)
			diffusivity = @(z) zeros(size(z));
		else
			diffusivity = diffusivities{m};
		end
		if(isa(diffusivity,'function_handle'))
			D = diffusivity(gridZ);	% evaluate diffusivity on gridZ
		else
			% diffusivity seems to be a single number
			D = diffusivity * ones(size(gridZ))
		end
		diffusive(m) = any(D); % check if diffusion is all-zero for this metabolite

		% get advection for this metabolite
		if(length(advections)==1)
			advection = advections{1};
		elseif(length(advections)==0)
			advection = @(z) zeros(size(z));
		else
			advection = advections{m};
		end
		if(isa(advection,'function_handle'))
			adv = advection(gridZ);	% evaluate advection on gridZ
		else
			% advection seems to be a single number
			adv = advection * ones(size(gridZ));
		end
		
		% check if D & adv contains NaNs
		if(check_nans_in_input && any(isnan(D))) 
			errormsg = sprintf('Diffusivity for %s contains NaN in considered depth range',metabolite_names{m});
			precalculations_OK = false; break;
		end
		if(check_nans_in_input && diffusive(m) && any(D==0)) 
			errormsg = sprintf('Diffusivity for %s is zero at some depths but not everywhere, but should be either zero everywhere or zero nowhere',metabolite_names{m});
			precalculations_OK = false; break;
		end
		if(check_nans_in_input && any(isnan(adv))) 
			errormsg = sprintf('Advection for %s contains NaN in considered depth range',metabolite_names{m});
			precalculations_OK = false; break;
		end
		
		% pre-calculate some derived variables
		grid_diffusivities(m,:) = D; 	% store for latter re-use
		grid_advections(m,:)	= adv; 	% store for latter re-use
		if(diffusive(m))
			V  = antiderivative(gridZ, adv./D); 	% V(z) = int_A^z advection(x)/D(x) dx
			eV = exp(V);
			E  = antiderivative(gridZ, exp(-V).*W./D); % E(z) = int_A^z exp(-V(x))W(x)/D(x) dx
			Er = antiderivative_reverse(gridZ, exp(-V).*W./D); % Er(z) = int_z^B exp(-V(x))W(x)/D(x) dx = E(B)-E(z). This is numerically more stable than just subtracting E(B)-E(z), especially when the integrant decays very rapidly along z.
		end
		green_adv(m,:) = adv; green_E(m,:) = E; green_Er(m,:) = Er; green_eV(m,:) = eV; green_W(m,:) = W;
		if((length(added_production_rates)>0) && (isa(added_production_rates{m},'function_handle')))
			has_added_production_rates = true;
			f = added_production_rates{m}(gridZ);
			if(check_nans_in_input && any(isnan(f))) 
				errormsg = sprintf('Added_production_rate for %s includes NaN in considered depth range',metabolite_names{m});
				precalculations_OK = false; break;
			end
			F = antiderivative(gridZ, f./W); % F(z) = int_A^z f(x)/W(x) dx
			if(diffusive(m))
				H = antiderivative(gridZ, exp(-V).*F.*W./D); % E(z) = int_A^z exp(-V(x))*F(x)*W(x)/D(x) dx
			end
		else
			has_added_production_rates = false;
		end
		
		if(diffusive(m))
			% transport includes diffusion
			if(isnan(BCvaluesA(m)) || isnan(BCvaluesB(m)))
				errormsg = sprintf('Missing boundary condition value at top and/or bottom for metabolite %s (found NaN)',metabolite_names{m});
				precalculations_OK = false; break;
			elseif(BCtypesA(m)==0 && BCtypesB(m)==0)
				% both boundary conditions are Dirichlet
				if(BCvaluesA(m)==0)
					% eliminate zero BC values from formula if possible, to reduce risk of 0*Inf
					if(BCvaluesB(m)==0)
						concentrations_BC_part(m,:) = 0;
					else
						concentrations_BC_part(m,:) = exp(V-V(NZ)).*((BCvaluesB(m)/E(NZ))*E);
					end
				else
					concentrations_BC_part(m,:) = exp(V).*(BCvaluesA(m) + ((BCvaluesB(m)*exp(-V(NZ))-BCvaluesA(m))/E(NZ))*E);
				end
				if(has_added_production_rates)
					concentrations_added_production_part(m,:) = (exp(V).*E)*H(NZ)/E(NZ) - exp(V).*H;
				end
				for s=1:NZ
					leftx  = (1:s);
					rightx = (s:NZ);
					greens(m,s,leftx)  = eV(leftx).*(E(leftx)/(E(NZ)*W(s))) .* Er(s);
					greens(m,s,rightx) = eV(rightx).*(E(s)/(E(NZ)*W(s))) .* Er(rightx);
				end
			elseif(BCtypesA(m)==0 && BCtypesB(m)==1) 
				concentrations_BC_part(m,:) = exp(V).*(BCvaluesA(m) + BCvaluesB(m)*E);
				if(has_added_production_rates)
					concentrations_added_production_part(m,:) = exp(V).*(F(NZ)*E - H);
				end
				for s=1:NZ
					leftx  = (1:s);
					rightx = (s:NZ);
					greens(m,s,leftx)  = eV(leftx).*E(leftx)/W(s);
					greens(m,s,rightx) = eV(rightx).*E(s)/W(s);
				end
			
			elseif(BCtypesA(m)==1 && BCtypesB(m)==0) 
				concentrations_BC_part(m,:) = exp(V).*(BCvaluesA(m)*(E(NZ)-E) + BCvaluesB(m)*exp(-V(NZ)));
				if(has_added_production_rates)
					concentrations_added_production_part(m,:) = exp(V).*(H(NZ) - H);
				end
				for s=1:NZ
					leftx  = (1:s);
					rightx = (s:NZ);
					greens(m,s,leftx)  = eV(leftx).*Er(s)/W(s);
					greens(m,s,rightx) = eV(rightx).*Er(rightx)/W(s);
				end
			
			elseif(BCtypesA(m)==1 && BCtypesB(m)==1)
				errormsg = sprintf('At least one of the boundary conditions must be Dirichlet (found both Robin for metabolite %s)',metabolite_names{m});
				precalculations_OK = false; break;
			else
				errormsg = sprintf('Unknown boundary condition types %d and/or %d for metabolite %s',BCtypesA(m),BCtypesB(m),metabolite_names{m});
				precalculations_OK = false; break;
			end
			
		else
			% purely advective transport, so solution family is very different, and only one boundary condition is used
			if((BCtypesA(m)==0) && (isnan(BCtypesB(m)) || isnan(BCvaluesB(m))))
				% Dirichlet BC at top
				BCtypesB(m) = NaN; BCvaluesB(m) = NaN; % ensure bottom BC is disabled
				concentrations_BC_part(m,:) = (W./adv) * BCvaluesA(m) * adv(1);
				if(has_added_production_rates)
					concentrations_added_production_part(m,:) = (W./adv).*F;
				end
				for s=1:NZ
					leftx  = (1:s);
					rightx = (s:NZ);
					greens(m,s,leftx)  = 0;
					greens(m,s,rightx) = (1/W(s))*(W(rightx)./adv(rightx));
				end

			elseif((isnan(BCtypesA(m)) || isnan(BCvaluesA(m))) && (BCtypesB(m)==0))
				% Dirichlet BC at bottom
				BCtypesA(m) = NaN; BCvaluesA(m) = NaN; % ensure top BC is disabled
				concentrations_BC_part(m,:) = (W./adv) * BCvaluesB(m) * (adv(NZ)/W(NZ));
				if(has_added_production_rates)
					concentrations_added_production_part(m,:) = (W./adv).*(F-F(NZ));
				end
				for s=1:NZ
					leftx  = (1:s);
					rightx = (s:NZ);
					greens(m,s,leftx)  = -(1/W(s))*(W(leftx)./adv(leftx));
					greens(m,s,rightx) = 0;
				end

			elseif((BCtypesA(m)==1) && (isnan(BCtypesB(m)) || isnan(BCvaluesB(m))))
				% Robin BC at top
				BCtypesB(m) = NaN; BCvaluesB(m) = NaN; % ensure bottom BC is disabled
				concentrations_BC_part(m,:) = (W./adv) * BCvaluesA(m);
				if(has_added_production_rates)
					concentrations_added_production_part(m,:) = (W./adv).*F;
				end
				for s=1:NZ
					leftx  = (1:s);
					rightx = (s:NZ);
					greens(m,s,leftx)  = 0;
					greens(m,s,rightx) = (1/W(s))*(W(rightx)./adv(rightx));
				end

			elseif((isnan(BCtypesA(m)) || isnan(BCvaluesA(m))) && (BCtypesB(m)==1))
				% Robin BC at bottom
				BCtypesA(m) = NaN; BCvaluesA(m) = NaN; % ensure top BC is disabled
				concentrations_BC_part(m,:) = -(W./adv) * (BCvaluesB(m)/W(NZ));
				if(has_added_production_rates)
					concentrations_added_production_part(m,:) = (W./adv).*(F-F(NZ));
				end
				for s=1:NZ
					leftx  = (1:s);
					rightx = (s:NZ);
					greens(m,s,leftx)  = -(1/W(s))*(W(leftx)./adv(leftx));
					greens(m,s,rightx) = 0;
				end
				
			elseif(isnan(BCtypesA(m)) && isnan(BCtypesA(m)))
				errormsg = sprintf('Both boundary conditions for metabolite %s are unspecified (NaN), but need exactly one boundary condition',metabolite_names{m});
				precalculations_OK = false; break;
			elseif((~isnan(BCtypesA(m))) && (~isnan(BCtypesA(m))))
				errormsg = sprintf('Both boundary conditions for metabolite %s are specified, but need exactly one boundary condition since transport is purely advective',metabolite_names{m});
				precalculations_OK = false; break;
			end
		end
	end	
	if(~precalculations_OK)
		concentrations_BC_part = NaN; concentrations_added_production_part = NaN;
		return;
	end
	
	% combine BC part to explicitly added production part into reaction-less solution
	concentrations_reactionless = concentrations_BC_part + concentrations_added_production_part;
	
	% define functional for calculating steady-state concentrations, for a given set of hotspots and rates
	function [concentrations, concentrations_reaction_part] = get_concentrations(hotspots, rates)
		hotspots = reshape(hotspots,NH,1);
		hotspots_right = zeros(NH,1);
		for c=1:NH
			hotspot_right = find(gridZ >= hotspots(c), 1, 'first');
			if(length(hotspot_right)==0)
				hotspots_right(c) = NZ;
			else
				hotspots_right(c) = hotspot_right;
			end
		end
		hotspots_left = max(1,hotspots_right-1);
		lambdas = (hotspots-gridZ(hotspots_left))./(gridZ(hotspots_right)-gridZ(hotspots_left));
		lambdas(hotspots_left==hotspots_right)==0;
		
		% sum partial solutions (one per metabolite per hotspot) to get metabolite profiles
		% concentrations will be an NM x NZ matrix
		sources 					 = stoichiometry * rates; % sources[m,c] will be the production rate of metabolite m at hotspot c
		concentrations_reaction_part = zeros(NM,NZ);
		local_greens 				 = zeros(NM,NZ);
		for c=1:NH
			if(isnan(hotspots(c))) continue; end
			if(INTERPOLATE_LOCAL_GREENS_FUNCTIONS)
				% linearly interpolate greens function between left & right grid point
				% This may not work very well, because the linear interpolation of Greens functions for nearby sources is not a good approximation for the local Greens function; at least when evaluated between the left & right interpolation anchors
				local_greens = squeeze((1-lambdas(c))*greens(:,hotspots_left(c),:) + lambdas(c)*greens(:,hotspots_right(c),:));
			else
				% calculate Green's function for sources localized at this hotspot ("local Greens function")
				% using the auxiliary pre-calculated variables E, Er, eV
				leftx  		  = (1:hotspots_left(c));
				rightx 		  = (hotspots_right(c):NZ);
				E_at_hotspot  = (1-lambdas(c))*green_E(:,hotspots_left(c)) + lambdas(c)*green_E(:,hotspots_right(c));
				Er_at_hotspot = (1-lambdas(c))*green_Er(:,hotspots_left(c)) + lambdas(c)*green_Er(:,hotspots_right(c));
				W_at_hotspot  = (1-lambdas(c))*green_W(:,hotspots_left(c)) + lambdas(c)*green_W(:,hotspots_right(c));
				for m=1:NM
					if(diffusive(m))
						% transport includes diffusion
						if(BCtypesA(m)==0 && BCtypesB(m)==0)
							local_greens(m,leftx)  = green_eV(m,leftx).*(green_E(m,leftx)/green_E(m,NZ)) .* Er_at_hotspot(m) ./ W_at_hotspot;
							local_greens(m,rightx) = green_eV(m,rightx).*(E_at_hotspot(m)/green_E(m,NZ)) .* green_Er(m,rightx) ./ W_at_hotspot;
						elseif(BCtypesA(m)==0 && BCtypesB(m)==1) 
							local_greens(m,leftx)  = green_eV(m,leftx).*green_E(m,leftx)./W_at_hotspot;
							local_greens(m,rightx) = green_eV(m,rightx).*E_at_hotspot(m)./W_at_hotspot;
						elseif(BCtypesA(m)==1 && BCtypesB(m)==0) 
							local_greens(m,leftx)  = green_eV(m,leftx).*Er_at_hotspot(m)./W_at_hotspot;
							local_greens(m,rightx) = green_eV(m,rightx).*green_Er(m,rightx)./W_at_hotspot;
						end
					else
						% purely advective transport, so treat separately and use only one boundary condition
						if(BCtypesA(m)==0)
							local_greens(m,leftx)  = 0;
							local_greens(m,rightx) = (1/W_at_hotspot)*(green_W(m,rightx)./green_adv(m,rightx));
						elseif(BCtypesB(m)==0)
							local_greens(m,leftx)  = -(1/W_at_hotspot)*(green_W(m,leftx)./green_adv(m,leftx));
							local_greens(m,rightx) = 0;
						elseif(BCtypesA(m)==1)
							local_greens(m,leftx)  = 0;
							local_greens(m,rightx) = (1/W_at_hotspot)*(green_W(m,rightx)./green_adv(m,rightx));
						elseif(BCtypesB(m)==1)
							local_greens(m,leftx)  = -(1/W_at_hotspot)*(green_W(m,leftx)./green_adv(m,leftx));
							local_greens(m,rightx) = 0;				
						end
					end
				end
			end
			concentrations_reaction_part = concentrations_reaction_part + bsxfun(@times, local_greens, sources(:,c));
		end
		concentrations = concentrations_reactionless + concentrations_reaction_part;
	end
	
	
	% define functional for calculating overall deviation of concentrations from non-negativity, for a given choice of fitted params
	% this function is used by fmincon as a non-linear vectorial constraint, specifying that all concentrations must be non-negative
	% if negative_concentration_penalty~=Inf, then this function returns empty vectors; in that case, negative concentrations are explicitly penalized in the stress function
	function [negativities_ineq,negativities_eq] = get_negativities(fparams)
		negativities_ineq = [];
		negativities_eq	  = [];
		if(not(isinf(negative_concentration_penalty)))
			return;
		end
		params   = fixed_params; params(fparam_indices) = fparams; % combined free with fixed parameters
		hotspots = params(1:NH);
		rates    = max(0,reshape(params((NH+1):end),NRe,NH));
		[concentrations, ~] = get_concentrations(hotspots,rates);
		if(any(isnan(concentrations)))
			negativities_ineq = zeros(NM*NZ,1);
			return;
		end
		% Each metabolite at each location corresponds to a separate constraint
		%negativities_ineq = -reshape(bsxfun(@rdivide,concentrations,negativity_scales),[NM*NZ,1]); % linear penalty
		%negativities_ineq = reshape(min(0,bsxfun(@rdivide,concentrations,negativity_scales)).^2,[NM*NZ,1]); % quadratic penalty, only for negative concentrations
		negativities_ineq = -reshape(min(0,bsxfun(@rdivide,concentrations,negativity_scales)),[NM*NZ,1]); % linear penalty, only for negative concentrations
		
		% alternative:  quadratic penalty for negative concentrations, keep linear values for positive concentrations
		%negativities_ineq = -reshape(bsxfun(@rdivide,concentrations,negativity_scales),[NM*NZ,1]);
		%negativities_ineq(negativities_ineq>0) = negativities_ineq(negativities_ineq>0).^2;

		% alternative:  quadratic penalty, but keeping correct sign
		%negativities_ineq = -reshape(bsxfun(@rdivide,concentrations,negativity_scales),[NM*NZ,1]);
		%negativities_ineq(negativities_ineq>0) = negativities_ineq(negativities_ineq>0).^2;
		%negativities_ineq(negativities_ineq<0) = -(negativities_ineq(negativities_ineq<0).^2);

		% old code. alternative: summarize along z-axis
		%negativities_eq = power(mean(power(abs(min(0,bsxfun(@rdivide,concentrations,negativity_scales))),NEGATIVE_CONCENTRATION_PENALTY_POWER),2),1/NEGATIVE_CONCENTRATION_PENALTY_POWER);
	end
	
	
	% determine DeltaG scales based on reaction-less profiles
	if(strcmpi(penalize_inefficiencies_by,'reaction_DeltaG'))
		DeltaGs = reaction_DeltaG(reaction_stoichiometry, reaction_DeltaG0, grid_temperatures, concentrations_reactionless, concentration_epsilon);
		%DeltaG_scales = (1/NZ) * abs(naninfsum(-min(0,DeltaGs),2)); % variant 1. DeltaG_scales will be a 1D column-vector of size NR
		DeltaG_scales = (1/(NZ*NR)) * sum(abs(naninfsum(-min(0,DeltaGs),2))); % variant 2
		%DeltaG_scales = (1/(NZ*NR)) * naninfsum(-DeltaGs,0); % variant 3
		DeltaG_scales(DeltaG_scales==0) = abs(reaction_DeltaG0(DeltaG_scales==0));
		DeltaG_scales(DeltaG_scales==0) = 1;
	end

	% determine concentration scales based on reaction-less profiles
	concentration_scales = nanmean(abs(concentrations_reactionless),2);
	if(all(concentration_scales==0))
		concentration_scales = ones(size(concentration_scales));
	else
		concentration_scales(concentration_scales==0) = mean_above(concentration_scales,0,1);
	end

	% determine relative metabolite_weights (if not provided)
	if(length(metabolite_weights)<=1)
		metabolite_weights = ones(NM,1);
	else
		nans = find(isnan(metabolite_weights));
		if(length(nans)==length(metabolite_weights))
			metabolite_weights = ones(NM,1);
		else
			metabolite_weights(nans) = nanmean(metabolite_weights);
		end
	end
	metabolite_weights = (metabolite_weights./sum(metabolite_weights)); % normalize metabolite_weights so that they become relative
	
	% define positivity scales if not provided
	if((length(positivity_scales)==0) || (all(isnan(positivity_scales))))
		positivity_scales = concentration_scales;
	elseif(length(positivity_scales)==1)
		positivity_scales = repmat(positivity_scales,NM,1);
	else
		nans = find(isnan(positivity_scales));
		positivity_scales(nans) = concentration_scales(nans);
		positivity_scales(positivity_scales==0) = mean_above(positivity_scales,0,1);
	end
		
	% define negativity scales if not provided
	if((length(negativity_scales)==0) || (all(isnan(negativity_scales))))
		negativity_scales = concentration_scales;
	elseif(length(negativity_scales)==1)
		negativity_scales = repmat(negativity_scales,NM,1);
	else
		nans = find(isnan(negativity_scales));
		negativity_scales(nans) = concentration_scales(nans);
		negativity_scales(negativity_scales==0) = mean_above(negativity_scales,0,1);
	end
	
	% estimate boundary influxes into the system, in the absence of any reactions
	% at reaction-less steady state these will be equal to the boundary outfluxes
	reactionless_influxes = zeros(NM,1);
	for m=1:NM
		if(diffusive(m))
			if((BCtypesA(m)==0) && (BCvaluesA(m)==0) && (BCtypesB(m)==0) && (BCvaluesB(m)==0))
				% zero-Dirichlet BCs at both ends, so zero flux
				reactionless_influxes(m) = 0;
			elseif((BCtypesA(m)==0) && (BCtypesB(m)==0))
				% Dirichlet BCs on both ends.
				reactionless_influxes(m) = abs(-((concentrations_reactionless(m,2)-concentrations_reactionless(m,1))/(gridZ(2)-gridZ(1)))*grid_diffusivities(m,1) + concentrations_reactionless(m,1)*grid_advections(m,1));
			elseif((BCtypesA(m)==1) && (BCvaluesA(m)~=0))
				reactionless_influxes(m) = abs(BCvaluesA(m));
			elseif((BCtypesB(m)==1) && (BCvaluesB(m)~=0))
				reactionless_influxes(m) = abs(BCvaluesB(m));
			else
				% at least one boundary has zero-flux BC, so at steady state there will be no flux across any boundary
				reactionless_influxes(m) = 0;
			end
		else
			% purely advective transport
			if(BCtypesA(m)==1)
				% Robin BC at top
				reactionless_influxes(m) = abs(BCvaluesA(m));
			elseif(BCtypesB(m)==1)
				% Robin BC at bottom
				reactionless_influxes(m) = abs(BCvaluesB(m));
			elseif(BCtypesA(m)==0)
				% Dirichlet BC at top
				reactionless_influxes(m) = abs(concentrations_reactionless(m,1)*grid_advections(m,1));
			elseif(BCtypesB(m)==0)
				% Dirichlet BC at bottom
				reactionless_influxes(m) = abs(concentrations_reactionless(m,NZ)*grid_advections(m,NZ));
			else
				reactionless_influxes(m) = 0;
			end
				
		end
	end
			
	% guess metabolite flux scales
	flux_scales = abs(reactionless_influxes);
	if((length(start_reaction_rates)>0) & not(any(isnan(start_reaction_rates))))
		start_flux_scales = mean(abs(reaction_stoichiometry * start_reaction_rates),2);
		flux_scales((flux_scales==0) & (start_flux_scales~=0)) = abs(start_flux_scales((flux_scales==0) & (start_flux_scales~=0)));
	end
	flux_epsilon = 1e-8 * mean_above(flux_scales,0,1);
	flux_scales(flux_scales<=flux_epsilon) = mean_above(flux_scales, flux_epsilon, 1);

	% guess typical reaction-rate scales (needed for normalizing parameters during fitting)
	rate_scales = repmat(abs(pinv(stoichiometry)*reactionless_influxes),1,NH);
	if(length(start_rates)>0)
		% use start_reaction_rates for scales, whenever they are provided and non-zero
		valids = (not(isnan(start_rates)) & (start_rates~=0));
		rate_scales(valids) = abs(start_rates(valids));
	end
	if(length(fixed_rates)>0)
		% use fixed_reaction_rates for scales, whenever they are available and non-zero
		valids = (not(isnan(fixed_rates)) & (fixed_rates~=0));
		rate_scales(valids) = abs(fixed_rates(valids));
	end
	rate_epsilon = 1e-8 * mean_above(rate_scales,0,1);
	rate_scales(rate_scales<=rate_epsilon) = mean_above(rate_scales, rate_epsilon,1);
	
	
	% determine parameters to be fixed vs. fitted
	if(length(fixed_hotspots)==0) fixed_hotspots = repmat(NaN,NH,1); end
	if(length(fixed_rates)==0) fixed_rates = repmat(NaN,NRe,NH); end
	fixed_params 	= [fixed_hotspots; reshape(fixed_rates,NRe*NH,1)];	% fixed parameter values and NaNs (for fitted parameters)
	fparam_indices 	= find(isnan(fixed_params)); 	% indices of fitted (free) parameters
	NFP 			= length(fparam_indices);		% number of fitted parameters
	NP				= length(fixed_params); 		% total number of parameters (fixed or fitted)
	
	% determine start params
	if(length(start_hotspots)==0) start_hotspots = repmat(NaN,NH,1); end
	if(length(start_rates)==0) start_rates = repmat(NaN,NRe,NH); end
	provided_start_params = [start_hotspots; reshape(start_rates,[NRe*NH 1])];
	provided_start_params(not(isnan(fixed_params))) = fixed_params(not(isnan(fixed_params))); % always apply fixed values to start params
	if(length(always_use_start_hotspots)==0) always_use_start_hotspots = zeros(NH,1); end
	if(length(always_use_start_rates)==0) always_use_start_rates = zeros(NRe,NH); end
	always_use_start_params = [always_use_start_hotspots; reshape(always_use_start_rates,NRe*NH,1)];
	% At this point, provided_start_params[] and always_use_start_params[] are guaranteed to be defined and full-length.
	% provided_start_params[] may include NaNs.
	% Whenever a param is fixed, its fixed value should also be applied to its start value
	
	
	% define objective function to be minimized ("stress")
	% params = [hotspots; reshape(rates,NRe*NH,1)]
	% hotspots: 1D array of size NH, specifying the positions of hotspots ("hotspots")
	% rates:   2D matrix of size NRe x NH, specifying the rate of each reaction at each hotspot
	function stress = get_stress(fparams)
		params   = fixed_params; params(fparam_indices) = fparams; % combined free with fixed parameters
		hotspots = params(1:NH);
		rates    = max(0,reshape(params((NH+1):end),NRe,NH));
		[concentrations, ~] = get_concentrations(hotspots,rates);
		if(any(isnan(concentrations)))
			stress = NaN;
			return;
		end
		stress = 0;
		% calculate penalty for negatives
		if((negative_concentration_penalty~=0) & not(isinf(negative_concentration_penalty)))
			stress = stress + negative_concentration_penalty*mean(power(mean(power(abs(min(0,bsxfun(@times,bsxfun(@rdivide,concentrations,negativity_scales),metabolite_weights))),NEGATIVE_CONCENTRATION_PENALTY_POWER),2),1/NEGATIVE_CONCENTRATION_PENALTY_POWER));
		end
		if(negative_flux_penalty~=0)
			net_sources 				= stoichiometry * rates; % net_sources[m,c] will be the production rate of metabolite m at hotspot c
			violating_sinks				= bsxfun(@rdivide,max(0,-net_sources),flux_scales);
			concentrations_at_hotspots 	= interp1(gridZ,concentrations',hotspots)'; % concentrations_at_hotspots[m,c] will be the concentration of metabolite m at hotspot c
			violating_sinks(concentrations_at_hotspots>0) = 0; % don't penalize net sinks when concentrations are strictly positive			
			stress = stress + negative_flux_penalty*mean(power(mean(power(bsxfun(@times,violating_sinks,metabolite_weights),NEGATIVE_FLUX_PENALTY_POWER),2),1/NEGATIVE_FLUX_PENALTY_POWER));
		end
		% calculate penalty for inefficiencies
		if(strcmpi(penalize_inefficiencies_by,'reaction_DeltaG'))
			DeltaGs = reaction_DeltaG(reaction_stoichiometry, reaction_DeltaG0, grid_temperatures, concentrations, concentration_epsilon);
			DeltaGstress = (1/(NZ*NR)) * sum(naninfsum(-min(0,DeltaGs),2)./DeltaG_scales); % variant 1. each reaction DeltaG normalized separately
			%DeltaGstress = (1/(NZ*NR)) * sum(naninfsum(-min(0,DeltaGs),2))/DeltaG_scales; % variant 2
			%DeltaGstress = exp((1/(NZ*NR)) * naninfsum(-DeltaGs,0)/DeltaG_scales-1); % variant 3
			stress  = stress + DeltaGstress;
		elseif(strcmpi(penalize_inefficiencies_by,'min_substrate_concentrations'))
			relative_concentrations = bsxfun(@rdivide,concentrations,positivity_scales);
			Cstress = 0;
			for r=1:NRe
				limiting_metabolites = find(stoichiometry(:,r)<0);
				if(length(limiting_metabolites)>0)
					Cstress = Cstress + power(mean(power(max(0,min(relative_concentrations(limiting_metabolites,:))),INEFFICIENCY_PENALTY_POWER)),1.0/INEFFICIENCY_PENALTY_POWER);
				end
			end
			stress = stress + Cstress/NRe;
		elseif(strcmpi(penalize_inefficiencies_by,'substrate_product'))
			relative_concentrations = bsxfun(@rdivide,concentrations,positivity_scales);
			Cstress = 0;
			for r=1:NRe
				limiting_metabolites = find(stoichiometry(:,r)<0);
				if(length(limiting_metabolites)>0)
					Cstress = Cstress + power(mean(power(prod(max(0,relative_concentrations(limiting_metabolites,:)),1),INEFFICIENCY_PENALTY_POWER)), 1.0/INEFFICIENCY_PENALTY_POWER);
				end
			end
			stress = stress + Cstress/NRe;
		elseif(strcmpi(penalize_inefficiencies_by,'substrate_geomean'))
			relative_concentrations = bsxfun(@rdivide,concentrations,positivity_scales);
			Cstress = 0;
			for r=1:NRe
				limiting_metabolites = find(stoichiometry(:,r)<0);
				if(length(limiting_metabolites)>0)
					Cstress = Cstress + power(mean(power(geomean(max(0,relative_concentrations(limiting_metabolites,:)),1),INEFFICIENCY_PENALTY_POWER)), 1.0/INEFFICIENCY_PENALTY_POWER);
				end
			end
			stress = stress + Cstress/NRe;
		end
		% add squared deviations from scatter data (if available)
		if(include_scatter_data)
			Dstress = 0;
			total_weight = 0;
			for m=1:NM
				if(length(parsed_concentration_scatter_data{m})>0)
					rights  = parsed_concentration_scatter_data{m}(:,1);
					lefts   = max(1,rights-1);
					rightX 	= gridZ(rights);
					leftX  	= gridZ(lefts);
					dataX	= parsed_concentration_scatter_data{m}(:,2);
					lambdas	= reshape((dataX-leftX)./(rightX-leftX), [1,length(leftX)]); % convex combination parameter for interpolating model onto data points
					lambdas(rights==lefts) = 0;
					model_values 	= reshape((lambdas.*concentrations(m,rights) + (1-lambdas).*concentrations(m,lefts)), length(rights),1);
					Dstress 		= Dstress + metabolite_weights(m) * sqrt(mean(((model_values-parsed_concentration_scatter_data{m}(:,3))./concentration_scales(m)).^2));
					total_weight	= total_weight + metabolite_weights(m);
				end
			end
			stress = stress + Dstress/total_weight;
		end
		% add squared deviations from grid data (if available)
		if(include_grid_data)
			Dstress = 0;
			total_weight = 0;
			for m=1:NM
				if(length(concentration_grid_data{m})>0)
					valids 			= find(not(isnan(concentration_grid_data{m})));
					Dstress			= Dstress + metabolite_weights(m) * sqrt(mean(((concentrations(m,valids)-concentration_grid_data{m}(valids))./concentration_scales(m)).^2));
					total_weight	= total_weight + metabolite_weights(m);
				end
			end
			stress = stress + Dstress/total_weight;
		end
	end

	% get typical stress scales
	if(strcmpi(penalize_inefficiencies_by,'reaction_DeltaG'))
		stress_scales = 1;
	elseif(strcmpi(penalize_inefficiencies_by,'min_substrate_concentrations'))
		stress_scales = nanmean(power(nanmean(power(bsxfun(@rdivide,concentrations_reactionless,positivity_scales),INEFFICIENCY_PENALTY_POWER),2),1/INEFFICIENCY_PENALTY_POWER));
	elseif(strcmpi(penalize_inefficiencies_by,'substrate_product'))
		stress_scales = nanmean(nanmean(bsxfun(@rdivide,concentrations_reactionless,positivity_scales),2));
	elseif(strcmpi(penalize_inefficiencies_by,'substrate_geomean'))
		stress_scales = nanmean(nanmean(bsxfun(@rdivide,concentrations_reactionless,positivity_scales),2));
	else
		stress_scales = 0;
	end
	if(include_scatter_data)
		Dstress = 0;
		for m=1:NM
			if(length(parsed_concentration_scatter_data{m})>0)
				Dstress = Dstress + std(metabolite_weights(m).*abs(parsed_concentration_scatter_data{m}(:,3)./concentration_scales(m)));
			end
		end
		stress_scales = stress_scales + Dstress/NMscatter_data;
	end
	if(include_grid_data)
		Dstress = 0;
		for m=1:NM
			if(length(concentration_grid_data{m})>0)
				Dstress = Dstress + nanstd(metabolite_weights(m).*abs(concentration_grid_data{m}./concentration_scales(m)));
			end
		end
		stress_scales = stress_scales + Dstress/NMgrid_data;
	end
	if(stress_scales<=0) stress_scales = 1.0; end
	
	% prepare inequality matrix to ensure fitted hotspots are always in ascending order
	% inequalities are encoded as: A * x <= b, where x are the fitted parameters
	% every inequality will be a separate column in this matrix
	% if there are NFH fitted hotspots (in ascending order), then there are NFH-1 inequalities
	NFH = sum(isnan(fixed_params(1:NH))); % number of fitted hotspots
	if(NFH<=1)
		inequalities_A = [];
		inequalities_b = [];
	else
		inequalities_b = zeros(NFH-1,1);
		inequalities_A = zeros(NFH-1,NFP);
		for i=1:(NFH-1)
			% encode inequality: fitted_hotspot(i)-fitted_hotspot(i+1) <= 0
			inequalities_A(i,i)   = 1;
			inequalities_A(i,i+1) = -1;
		end
	end

	% allocate space for fitting results across trials
	hotspots_all_trials			= NaN*zeros(Ntrials, NH);
	rates_all_trials 			= NaN*zeros(Ntrials, NRe, NH);
	reaction_rates_all_trials 	= NaN*zeros(Ntrials, NR, NH);
	stress_all_trials 			= repmat(NaN,Ntrials,1);
	R2s_grid_data_all_trials	= repmat(NaN,Ntrials,NM);
	R2s_scatter_data_all_trials	= repmat(NaN,Ntrials,NM);
	exit_flags_all_trials		= repmat(NaN,Ntrials,1);
	Niterations_all_trials		= repmat(NaN,Ntrials,1);
	Nevaluations_all_trials		= repmat(NaN,Ntrials,1);

	if(NH==0)
		best_hotspots 	= [];
		best_rates   	= zeros([NRe 0]);
		best_stress 	= get_stress([]);
		best_exit_flag	= NaN;
		best_Niterations= 0;

	elseif(NFP==0)
		if(verbose) fprintf('%sNote: All %d model parameters are fixed\n',verbose_prefix,NP); end
		best_params		= fixed_params;
		best_hotspots	= best_params(1:NH);
		best_rates 		= max(0,reshape(best_params((NH+1):end),NRe,NH));
		best_stress		= get_stress([]);
		best_exit_flag	= NaN;
		best_Niterations= 0;

	else
		if(verbose) fprintf('%sRunning SMF analysis with %d hotspots..\n',verbose_prefix,NH); end
		if(verbose) fprintf('%sNote: Estimating %d out of %d model parameters..\n',verbose_prefix,NFP,NP); end
		best_hotspots 	= NaN;
		best_rates 	 	= NaN;
		best_stress 	= NaN;
		best_exit_flag	= NaN;
		
		% determine bounds & scales & shifts for all parameters
		param_scales = [repmat(zB-zA,NH,1); reshape(rate_scales,NRe*NH,1)];
		if(length(start_rates)>0)
			% use provided starting params to define param scales, whenever the former are non-zero and non-NaN
			provided_start_scales 			 = [repmat(zB-zA,NH,1); reshape(abs(start_rates),[NRe*NH 1])];
			valid_start_scales 				 = (not(isnan(provided_start_scales)) & (provided_start_scales~=0));
			param_scales(valid_start_scales) = provided_start_scales(valid_start_scales);
		end
		param_scales(param_scales==0) = 1;
		min_rates	 = zeros(NRe,NH);
		max_rates	 = repmat(Inf,NRe,NH);
		min_params	 = [repmat(allowed_hotspot_range(1),NH,1); reshape(min_rates,NRe*NH,1)];
		max_params	 = [repmat(allowed_hotspot_range(2),NH,1); reshape(max_rates,NRe*NH,1)];
		param_shifts = [repmat(zA,NH,1); zeros(NRe*NH,1)];
		% refine hotspot bounds to account for fixed hotspots and ensure ascending order
		for c=1:NH
			if(not(isnan(fixed_params(c))))
				min_params(c) = fixed_params(c);
			elseif(c>1)
				min_params(c) = min_params(c-1);
			end
		end
		for c=NH:-1:1
			if(not(isnan(fixed_params(c))))
				max_params(c) = fixed_params(c);
			elseif(c<NH)
				max_params(c) = max_params(c+1);
			end
		end

		% constrain bounds & scales & shifts to just fitted params
		fparam_shifts = param_shifts(fparam_indices);
		fparam_scales = param_scales(fparam_indices);
		min_fparams	  = min_params(fparam_indices);
		max_fparams	  = max_params(fparam_indices);

		% Technical options for optimization algorithm
		% Common algorithm choices are 'interior-point' or 'sqp'. 
		%   For large models, interior-point seems to be faster than sqp.
		%	However, in the presence of non-linear constraints, interior-point often returns a worse point than the initial guess.
		% 	Interior-point is also less robust than SQP for non-convex problems.
		%   If all hotspots are fixed, and only reaction rates are fitted, the constraints become linear, in which case interior-point may be fine (this has not yet been thoroughly tested)
		%	Hence, sqp may be a slower but more reliable choice.
		%	For a comparison of algorithms see: https://www.mathworks.com/help/optim/ug/choosing-the-algorithm.html
		if(strcmp(optim_method,'fmincon'))
			fit_options = optimset(	'Algorithm', 'interior-point',... 
									'Hessian', 'bfgs',...
									'MaxIter',optim_max_iter,...
									'MaxFunEvals',optim_max_evals,...
									'Display','off',... % set to 'iter' for debugging, 'off' for normal mode
									'FunValCheck','off',...
									'PlotFcns',[],... % set to {'optimplotfval','optimplotx','optimplotstepsize','optimplotconstrviolation'} for debugging in non-parallel mode, or to [] otherwise.
									'FinDiffType','central',...
									'TolFun', objective_tolerance*stress_scales, ...
									'TolX', parameter_tolerance, ...
									'TolCon', 1e-8,...
									'AlwaysHonorConstraints', 'bounds',...
									'InitBarrierParam',0.1,...
									'DiffMinChange',1e-7,...
									'DiffMaxChange',1);
		elseif(strcmp(optim_method,'patternsearch'))
			% Note: Newer versions of MATLAB (>2015b) may require using optimoptions() instead of psoptimset()
			fit_options = psoptimset(	'MaxIter',optim_max_iter,...
										'MaxFunEvals',optim_max_evals,...
										'Display','off',... % set to 'iter' for debugging, 'off' for normal mode
										'PlotFcns',[],... % set to {'psplotbestf','psplotbestx','psplotmeshsize'} for debugging in non-parallel mode, or to [] otherwise.
										'TolFun', objective_tolerance*stress_scales, ...
										'TolX', parameter_tolerance, ...
										'TolCon', 1e-8,...
										'TolMesh', 1e-8,...
										'PollMethod', 'GSSPositiveBasis2N',...
										'SearchMethod', 'MADSPositiveBasis2N', ...
										'Vectorized', 'off',... % objective function and constraints are not vectorized
										'UseParallel', false,...
										'InitialMeshSize', 1.0,...
										'MeshExpansion', 2.0,...
										'SearchMethod', 'GPSPositiveBasis2N',...
										'CompletePoll','on',... %  search poll the entire mesh at each iteration, to reduce risk of local minima
										'ScaleMesh', 'off');
		end

		% run fitting trials in parallel
		concentration_function_handle 	= @get_concentrations; % need to refer to get_concentrations function via a handle in parfor loop
		stress_function_handle 			= @get_stress; % need to refer to stress function via a handle in parfor loop
		negativities_function_handle 	= @get_negativities;
		for (round=1:Nrounds)
			if((round>1) && verbose) fprintf('%s  Note: Repeating all fitting trials because previous rounds failed. Starting round #%d..\n',verbose_prefix,round); end
			% re-initialize storage
			hotspots_all_trials(:) 			= NaN;
			rates_all_trials(:) 			= NaN;
			reaction_rates_all_trials(:) 	= NaN;
			stress_all_trials(:) 			= NaN;
			R2s_grid_data_all_trials(:) 	= NaN;
			R2s_scatter_data_all_trials(:) 	= NaN;
			exit_flags_all_trials(:) 		= NaN;
			Niterations_all_trials(:) 		= NaN;
			Nevaluations_all_trials(:)		= NaN;
			best_hotspots 	= NaN;
			best_rates 	 	= NaN;
			best_stress 	= NaN;
			best_exit_flag	= NaN;
			
			%warning('off','MATLAB:mir_warning_maybe_uninitialized_temporary'); % suppress this annoying warning about temporary variables inside the parfor loop
			parfor (trial=1:Ntrials, ternary(parallel,Inf,0))
			%for trial=1:Ntrials % debug
				if(verbose & not(parallel)) fprintf('%s  Trial %d..',verbose_prefix,trial); end
				% initialize some loop-internal variables, to silence some annoying warnings by MATLAB
				scaled_fitted_fparams = NaN; trial_stress = NaN; trial_exit_flag = NaN; trial_output = NaN;
				
				% choose start params: [start_hotspots; start_rates_at_each_hotspot]
				% some start points may be invalid (e.g., violate constraints or yield a NaN objective value), so keep generating random start points if needed until a valid one is found
				% Note that the positivity constraint (i.e., all concentrations are non-negative) may in principle be violated by the start point, since fmincon seems to have a way to recover from these; however, if possible, try to find a start point that does not violate positivity constraints in the first place
				start_fparams = [];
				for a=1:Nstart_attempts
					%candidate_start_params = [repmat(NaN,NH,1); reshape(rate_scales,NRe*NH,1).*(10.^(max_random_rate_factor*(0.5-rand(NRe*NH,1))))]; % randomly vary start rates on a logarithmic scale
					candidate_start_params = [repmat(NaN,NH,1); reshape(randomize_symmetric(rate_scales, min_rates, min(max_rates,rate_scales*max_random_rate_factor)),NRe*NH,1)]; % randomly vary start rates on a linear scale
					
					if((a==1) && (trial==1))
						% use provided start params (if available)
						provided_start_params_to_use = not(isnan(provided_start_params));
						candidate_start_params(provided_start_params_to_use) = provided_start_params(provided_start_params_to_use);
					elseif(length(provided_start_params)>0)
						% only use provided start params that are to be used in all fitting trials
						provided_start_params_to_use = (always_use_start_params & not(isnan(provided_start_params)));
						candidate_start_params(provided_start_params_to_use) = provided_start_params(provided_start_params_to_use);
					end
					%randomly choose non-provided (missing) start hotspots, always ensuring ascending order
					maxzs = candidate_start_params(1:NH);
					for c=NH:-1:1
						if(isnan(maxzs(c)))
							if(c==NH) maxzs(c) = allowed_hotspot_range(2); else maxzs(c) = maxzs(c+1); end
						end
					end
					for c=1:NH
						if(isnan(candidate_start_params(c)))
							if(c==1) minz = allowed_hotspot_range(1); else minz = candidate_start_params(c-1); end
							candidate_start_params(c) = minz + rand(1,1)*(maxzs(c)-minz);
						end
					end
					% constrain start params within bounds just in case
					candidate_start_params  = min(max_params,max(min_params,candidate_start_params));
					% extract start values for just fitted parameters
					candidate_start_fparams = candidate_start_params(fparam_indices);
										
					% check if objective is defined for the start params
					% this is a potentially fatal checkpoint, so apply first
					start_objective_value = feval(stress_function_handle, candidate_start_fparams);
					if(isnan(start_objective_value) | isinf(start_objective_value))
						% this seems to be an invalid start point, so reject
						continue;
					end
					
					% fatal checkpoints passed, now proceed with non-fatal checkpoints (i.e., start params may be used if needed)
					start_fparams = candidate_start_fparams;
					
					% check if this start point satisfies non-linear constraints
					% make an exception if this is the first trial and we're using the starting params provided by the user, and if the latter are a complete set of params (none missing)
					if(isinf(negative_concentration_penalty))
						[negativities_ineq,negativities_eq] = feval(negativities_function_handle, candidate_start_fparams);
						if(any(negativities_ineq>0) || any(negativities_eq>0))
							% this seems to be an invalid starting point
							if((a==1) && (trial==1) && all(not(isnan(provided_start_params))))
								% start point was provided by user, so accept it anyway
								if(verbose) fprintf('%s  WARNING: Provided start point results in negative concentrations; proceeding anyway\n',verbose_prefix); end
							else
								% try to find a better start point
								continue;
							end
						end
					end
					% start point passed all tests, so accept, and proceed with fitting
					break;
				end
				
				if(length(start_fparams)==0)
					% no valid start params found, skip trial
					if(verbose & not(parallel)) fprintf('%s  Failed trial %d: Could not find valid start params\n',verbose_prefix,trial); end
					exit_flags_all_trials(trial) = -1;
					continue;
				end

				% calculate objective for start point, to check afterwards if fmincon actually found a better point
				start_stress = feval(stress_function_handle, start_fparams);
						
				% run fitting algorithm
				warning('off','MATLAB:nearlySingularMatrix'); 	% turn off this annoying warning
				warning('off','MATLAB:SingularMatrix'); 		% turn off this annoying warning
				warning('off','MATLAB:rankDeficientMatrix'); 	% turn off this annoying warning
				warning('off','MATLAB:illConditionedMatrix'); 	% turn off this annoying warning
				try
					if(strcmp(optim_method,'fmincon'))
						[scaled_fitted_fparams,...
						trial_stress,...
						trial_exit_flag,...
						trial_output] = fmincon(@(fparams) feval(stress_function_handle,(fparams.*fparam_scales)+fparam_shifts), ...
												(start_fparams-fparam_shifts) ./ fparam_scales, ...
												inequalities_A, inequalities_b, ...
												[], [], ...
												(min_fparams-fparam_shifts) ./ fparam_scales, ...
												(max_fparams-fparam_shifts) ./ fparam_scales, ...
												@(fparams) feval(negativities_function_handle,(fparams.*fparam_scales)+fparam_shifts), ...
												fit_options);
					elseif(strcmp(optim_method,'patternsearch'))
						[scaled_fitted_fparams,...
						trial_stress,...
						trial_exit_flag,...
						trial_output] = patternsearch(	@(fparams) feval(stress_function_handle,(fparams.*fparam_scales)+fparam_shifts), ...
														(start_fparams-fparam_shifts) ./ fparam_scales, ...
														inequalities_A, inequalities_b, ...
														[], [], ...
														(min_fparams-fparam_shifts) ./ fparam_scales, ...
														(max_fparams-fparam_shifts) ./ fparam_scales, ...
														@(fparams) feval(negativities_function_handle,(fparams.*fparam_scales)+fparam_shifts), ...
														fit_options);
					end
				catch exception
					% an error occurred during fitting
					if(verbose) fprintf('%s  WARNING: Failed trial %d: %s threw an exception: %s\n',verbose_prefix,trial,optim_method,strrep(getReport(exception,'basic'),sprintf('\n'),' ')); end
					exit_flags_all_trials(trial) = -1;
					continue;
				end
				
				% check if start point was better than final point (this can happen sometimes)
				% in that case, and only if negativity constraints are satisfied, replace the fmincon output with a dummy output representing the start point
				if(isnan(trial_stress) | (start_stress<trial_stress))
					[negativities_ineq,negativities_eq] = feval(negativities_function_handle, start_fparams);
					if(not(isinf(negative_concentration_penalty)) || (all(negativities_ineq<=0) && all(negativities_eq<=0)))
						trial_stress 			= start_stress;
						trial_exit_flag 		= 10;
						trial_output.iterations = 1;
						if(strcmp(optim_method,'fmincon')) trial_output.funcCount=1; else trial_output.funccount=1; end
						scaled_fitted_fparams	= (start_fparams-fparam_shifts) ./ fparam_scales;
					end
				end
				
				if(verbose & not(parallel)) fprintf(' stress=%g, exit_flag=%d, Niterations=%d\n',trial_stress,trial_exit_flag,trial_output.iterations); end

				% parse & record fitting result
				% if trial_exit_flag==0, then fitting did not converge, so be more careful when accepting solutions in that case
				exit_flags_all_trials(trial)  = trial_exit_flag;
				Niterations_all_trials(trial) = trial_output.iterations;
				if(strcmp(optim_method,'fmincon')) Nevaluations_all_trials(trial)= trial_output.funcCount; else Nevaluations_all_trials(trial)= trial_output.funccount; end
				if((trial_exit_flag==0) && isinf(negative_concentration_penalty))
					% fitting did not converge, so only accept final point if it at least satisfies non-linear constraints
					[negativities_ineq,negativities_eq] = feval(negativities_function_handle, (scaled_fitted_fparams.*fparam_scales)+fparam_shifts);
					if(any(negativities_ineq>0) || any(negativities_eq>0))
						% omit this solution, since it does not even satisfy constraints
						continue;
					end
				end
				if(trial_exit_flag>=0)
					fitted_fparams 	= (scaled_fitted_fparams .* fparam_scales) + fparam_shifts; % reverse parameter rescaling & shifting
					params 			= fixed_params; params(fparam_indices) = fitted_fparams; % combine fitted & fixed parameters
					stress_all_trials(trial) 	 		 = trial_stress;
					hotspots							 = params(1:NH);
					rates 								 = max(0,reshape(params((NH+1):end),NRe,NH));
					hotspots_all_trials(trial,:) 		 = hotspots;
					rates_all_trials(trial,:,:) 		 = rates;
					reaction_rates_all_trials(trial,:,:) = rates(1:NR,:);				
					% compare predictions of this trial with data
					if(R2s_for_all_trials & ((length(concentration_grid_data)>0) | (length(concentration_scatter_data)>0)))
						[concentrations, ~] 				 = feval(concentration_function_handle, hotspots, rates);
						R2s_grid_data_all_trials(trial,:)	 = get_griddata_R2s(gridZ,concentrations,concentration_grid_data);
						R2s_scatter_data_all_trials(trial,:) = get_scatterdata_R2s(gridZ,concentrations,concentration_scatter_data);
					end
				end
			end
			Nfailed_trials = Nfailed_trials + sum(exit_flags_all_trials<=0);
		
			% find best trial from this round
			[candidate_best_stress, candidate_best_trial] = nanmin(stress_all_trials);
			if(not(isnan(candidate_best_stress)))
				best_stress 	 = candidate_best_stress;
				best_exit_flag	 = exit_flags_all_trials(candidate_best_trial);
				best_Niterations = Niterations_all_trials(candidate_best_trial);
				best_hotspots 	 = hotspots_all_trials(candidate_best_trial,:);
				best_rates   	 = reshape(rates_all_trials(candidate_best_trial,:,:), [NRe NH]);
			end
			
			% check if this round failed, otherwise finish fitting
			if(isnan(best_stress) && verbose)
				fprintf('%s  WARNING: All fitting trials failed for round #%d\n',verbose_prefix,round); 
			else
				% check for spurious hotspots
				spurious_hotspots = find((best_hotspots<allowed_hotspot_range(1) + BOUNDARY_EPSILON_FOR_SPURIOUS_HOTSPOTS) | (best_hotspots>allowed_hotspot_range(2) - BOUNDARY_EPSILON_FOR_SPURIOUS_HOTSPOTS));
				if(length(spurious_hotspots)==0)
					% all looks good with this round, so finish fitting
					break
				elseif(verbose) 
					fprintf('%s  WARNING: %d fitted hotspots (located at depths %s m) are likely spurious, in round #%d (stress %g)\n',verbose_prefix,length(spurious_hotspots),sprint_numeric_vector(best_hotspots(spurious_hotspots),', '),round,best_stress);
				end
			end
			if((round==Nrounds) && (Nrounds>1) && verbose)
				fprintf('%s  WARNING: Reached max allowed number of fitting rounds, so using whatever solution I got from the last round\n',verbose_prefix);
			end
		end
	end

	% extract information from best fit
	if(isnan(best_stress))
		errormsg = sprintf('All fitting trials failed in all rounds'); 
	else
		stress  			= best_stress;
		exit_flag			= best_exit_flag;
		Niterations			= best_Niterations;
		hotspots 			= best_hotspots;
		rates 				= best_rates;
		
		% check for spurious hotspots
		spurious_hotspots = find((hotspots<allowed_hotspot_range(1) + BOUNDARY_EPSILON_FOR_SPURIOUS_HOTSPOTS) | (hotspots>allowed_hotspot_range(2) - BOUNDARY_EPSILON_FOR_SPURIOUS_HOTSPOTS));
		if(length(spurious_hotspots)>0)
			if(verbose) fprintf('%sWARNING: %d fitted hotspots (located at depths %s m) are likely spurious\n',verbose_prefix,length(spurious_hotspots),sprint_numeric_vector(hotspots(spurious_hotspots),', ')); end
			if(omit_spurious_hotspots)
				if(verbose) fprintf('%sNote: Omitting likely spurious hotspots (setting rates to zero)..\n',verbose_prefix); end
				rates(:,spurious_hotspots) = 0;
			end
		end
			
		% check for overlapping hotspots
		dz = mean(diff(gridZ));
		overlapping_hotspots = find(diff(hotspots)<dz);
		if((length(overlapping_hotspots)>0) && verbose)
			fprintf('%sNote: %d hotspots seem to be overlapping; perhaps fewer hotspots are needed for this model?\n',verbose_prefix,length(overlapping_hotspots)+1);
		end
	
		% calculate metabolic net source rates & concentrations based on final solution
		metabolite_net_source_rates = stoichiometry * rates;
		[concentrations, concentrations_reaction_part] = get_concentrations(hotspots, rates);
		
		% calculate metabolite net influxes through the top & bottom boundaries, based on the local concentrations & gradients
		top_influxes	 = -(concentrations(:,2)-concentrations(:,1))./(gridZ(2)-gridZ(1)) .* grid_diffusivities(:,1) + concentrations(:,1) .* grid_advections(:,1);
		bottom_influxes	 = (concentrations(:,end)-concentrations(:,end-1))./(gridZ(end)-gridZ(end-1)) .* grid_diffusivities(:,end) - concentrations(:,end) .* grid_advections(:,end);
		metabolite_boundary_influxes = [top_influxes bottom_influxes];
		
		% look for for negligible hotspots (i.e., where substrate fluxes are much smaller than in other hotspots)
		if(negligible_rel_flux_threshold>=0)
			max_abs_metabolite_net_source_rates = max(abs(metabolite_net_source_rates),[],2); % maximum flux magnitude for each metabolite (across hotspots)
			for c=1:NH
				if(omit_spurious_hotspots && ismember(c,spurious_hotspots)) continue; end % no need to classify a removed spurious hotspot as negligible
				if(all(abs(metabolite_net_source_rates(:,c))<=negligible_rel_flux_threshold * max_abs_metabolite_net_source_rates))
					negligible_hotspots(end+1) = c;
				end
			end
			if(length(negligible_hotspots)>0)
				fprintf('%sNote: Some hotspots (%s) appear to be negligible; perhaps fewer hotspots are needed for this model?\n',verbose_prefix,sprint_numeric_vector(negligible_hotspots,', '));
			end
		end
		
		% compare final predictions to available concentration data
		R2s_grid_data 	  = get_griddata_R2s(gridZ,concentrations,concentration_grid_data);
		R2s_scatter_data  = get_scatterdata_R2s(gridZ,concentrations,concentration_scatter_data);
			
		% split rates into reaction rates and ghost source/sink rates
		reaction_rates 		= rates(1:NR,:);
		ghost_source_rates 	= rates((NR+1):(NR+NGSources),:);
		ghost_sink_rates	= rates((NR+NGSources+1):NRe,:);
	end
end




% Ternary operator for C-like conditional expressions
function c = ternary(condition, a , b)
    if condition
        c = a;
    else
        c = b;
    end
end


function X2 = SQ(X)
	X2 = X.^2;
end


% returns Gibbs free energy of reactions
% functional_chemical{} is a cell array of spatiotemporal function handles
% Input:
%   stoichiometry: 2D array of size NM x NR, stoichiometric coefficients of reactions
%   DeltaG0: 1D vector of size NR or 2D array of size NR x NZ, standard Gibbs free energies of reactions. In the former case, DeltaG0 of a reaction is assumed to be the same for all locations.
%	T: 1D vector of size NZ, listing temperatures over NZ locations (in degrees C)
%   concentrations: 2D array of size NM x NZ, listing concentrations of NM metabolites over NZ locations
%   concentration_epsilon: single number, specifying the threshold below which metabolite concentrations are essentially zero. Any reactions whose substrate or product concentrations are below this threshold, will have DeltaG = Inf or -Inf, respectively.
% Output:
%   DeltaG: 2D array of size NR x NZ, listing Gibbs free energies of reactions
% By convention, DeltaG=Inf whenever a substrate has concentration <=0.
function DeltaG = reaction_DeltaG(stoichiometry, DeltaG0, T, concentrations, concentration_epsilon)
	Rg = 8.3144621e-3; 	% Gas constant (kJ/(K*mol))
	NZ = size(concentrations,2);
	NM = size(concentrations,1);
	NR = size(stoichiometry,2);
	if(size(DeltaG0,1)==1)
		DeltaG0 = DeltaG0'; 
	end
	if(size(DeltaG0,2)==1)
		DeltaG = repmat(reshape(DeltaG0,[NR 1]),1,NZ);
	else
		DeltaG = DeltaG0;
	end
	T = reshape(T,[1 NZ]); % make sure T is row vector
	for r=1:NR
		relevants   = find(stoichiometry(:,r)~=0);
		DeltaG(r,:) = DeltaG(r,:) + Rg * (T+273.15).*(stoichiometry(relevants,r)' * log(concentrations(relevants,:)));

		% set DeltaG of reactions to -Inf if one of the products has zero concentration
		products 		= find(stoichiometry(:,r)>0);
		zeros 	 		= any(concentrations(products,:)<=concentration_epsilon, 1);
		DeltaG(r,zeros) = -inf;

		% set DeltaG of reactions to +Inf if one of the substrates has zero concentration
		% this should come last, so that it overwrites other overlapping scenarios
		substrates 		= find(stoichiometry(:,r)<0);
		zeros 			= any(concentrations(substrates,:)<=concentration_epsilon, 1);
		DeltaG(r,zeros) = inf;
	end
end


% integrate y over its first dimension along x
% x is a vector of length L
% y is a vector of length L or a matrix of size L x M
function I = integrateVector(x, y)
	dx = diff(x);
	ym = (y(2:end,:)+y(1:end-1,:))/2;
	I = squeeze(sum(bsxfun(@times,ym,reshape(dx,length(dx),1)), 1));
end


% calculate the antiderivative of a scalar-valued function along a 1D axis
% return vector of size size(f), whose i-th element is int_{z[1]}^{z[i]}f(x) dx
% z and f should be vectors of equal size
% z should be incremental order
function A = antiderivative(z, f)
	NZ = length(z);
	if(any(not(size(z)==size(f)))) z=z'; end
	if(size(z,1)==1)
		A = [0 cumsum(diff(z).*(0.5*(f(1:(NZ-1))+f(2:NZ))))]; % treat as row vector
	else
		A = [0; cumsum(diff(z).*(0.5*(f(1:(NZ-1))+f(2:NZ))))]; % treat as column vector
	end	
end

% calculate the antiderivative of a scalar-valued function along a reversed 1D axis
% return vector of size size(f), whose i-th element is int_{z[i]}^{z[NZ]}f(x) dx = - int_{z[NZ]}^{z[i]}f(x) dx
% z and f should be 1D vectors of size NZ
% z should be incremental order
function A = antiderivative_reverse(z, f)
	NZ = length(z);
	if(any(not(size(z)==size(f)))) z=z'; end
	if(size(z,1)==1)
		A  = [cumsum(diff(z).*(0.5*(f(1:(NZ-1))+f(2:NZ))),'reverse') 0]; % treat as row vector
	else
		A  = [cumsum(diff(z).*(0.5*(f(1:(NZ-1))+f(2:NZ))),'reverse'); 0]; % treat as column vector
	end
end


% calculate sum of 2D array over a specific axis, ignoring NaN and Inf
% A: 2D array of size NR x NC
% axis: either 0, 1 or 2. If 0, then the sum of all array elements is returned as a single scalar.
% For example, if axis=1, then the returned array will be of size 1 x NC.
function S = naninfsum(A,axis)
	if(axis==0)
		S = naninfsum(reshape(A,[1,prod(size(A))]), 2);
	else
		if(axis==1) A=A'; end
		NR = size(A,1);
		S  = zeros(NR,1);
		for r=1:NR
			valids = not(isnan(A(r,:)) | isinf(A(r,:)));
			if(any(valids))
				S(r) = sum(A(r,valids));
			end
		end
		if(axis==1) S=S'; end
	end
end


% calculate the antiderivative of a vector-valued function along a 1D axis
% return 2D array of size length(z) x NM, whose i-th row is int_{z[1]}^{z[i]}f(x,:) dx
% z should be a 1D vector of size NZ
% f should be a 2D array of size NZ x NM
function A = antiderivative_of_matrix(z, f)
	NZ = length(z);
	NM = size(f,2);
	A  = [zeros(1,NM); cumsum(bsxfun(@times, reshape(diff(z),[NZ-1,1]), (0.5*(f(1:(NZ-1),:)+f(2:NZ,:)))), 1)];
end


% calculate the weighted geometric mean of a matrix along a specific dimension (dim=1 or 2)
% the optional weights are the exponents used for each component.
% e.g. the numbers X1,X2,X3 with weights 1 2 5 will have geometric mean = (X1^1 * X2^2 * X3^5)^(1/(1+2+5))
% The weights don't need to be normalized, but they should all be non-negative numbers and their sum should be non-zero.
% weights can also be an empty vector (all components weighted equal)
function G = weighted_geomean(X,dim,weights)
	if(dim==1)
		% averaging will always be done along axis 2
		X = X';
	end
	if(length(weights)==0)
		G = power(prod(X,2), 1.0/size(X,2));
	else
		if(size(weights,1)~=1)
			weights = weights'; % make sure weights are row-vector
		end
		powers 	= bsxfun(@power,X,weights);
		G 		= power(prod(powers,2),1/sum(weights));
	end
	if(dim==1)
		% revert to original orientation
		G = G';
	end
end



% calculate fractions of explained variance (R2) of depth profile data by model predictions
% input:
%	gridZ: 1D numerical vector of size NZ, listing depths
%	predictions: 2D numerical array of size NM x NZ, listing concentrations for NM metabolites along NZ depths
%	data: 1D cell array of size NM, each entry of which is a 1D vector of size NZ listing concentrations across depths, or 
function R2s = get_griddata_R2s(gridZ,predictions,data)
	NM  = size(predictions,1);
	NZ  = size(predictions,2);
	R2s = repmat(NaN,NM,1);
	if(length(data)==0)
		% no data available
		return;
	end
	for m=1:NM
		mdata = data{m};
		if(length(mdata)==0) continue; end % no data available for this metabolite
		valids = find(not(isnan(mdata) | isnan(predictions(m,:)))); % find depths with valid data & predictions
		if(length(valids)<=1) continue; end
		R2s(m) = 1.0 - sum((predictions(m,valids)-mdata(valids)).^2)./(length(valids)*var(mdata(valids)));
	end
end


% calculate fractions of explained variance (R2) of depth profile data by model predictions
% input:
%	gridZ: 1D numerical vector of size NZ, listing depths
%	predictions: 2D numerical array of size NM x NZ, listing concentrations for NM metabolites along NZ depths
%	data: 1D cell array of size NM, each entry of which is a 2D array of size NDm x 2 listing scattered concentration measurements (column 1: depth, column 2: concentration)
function R2s = get_scatterdata_R2s(gridZ,predictions,data)
	NM  = size(predictions,1);
	NZ  = size(predictions,2);
	R2s = repmat(NaN,NM,1);
	if(length(data)==0)
		% no data available
		return;
	end
	for m=1:NM
		mdata = data{m};
		if((size(mdata,1)==0) | (size(mdata,1)<2)) continue; end % no data available for this metabolite
		mdata = mdata(not(isnan(mdata(:,1)) | isnan(mdata(:,2)) | (mdata(:,1)<gridZ(1)) | (mdata(:,1)>gridZ(end))),:); % remove NaN scatterpoints and scatterpoints outside of gridZ
		if(size(mdata,1)==0) continue; end % all data filtered out
		mpredictions = interp1(gridZ,predictions(m,:),mdata(:,1),'linear',NaN); % interpolate model predictions onto scattered data points
		valids = find(not(isnan(mpredictions))); % find scatterpoints for which predictions were possible
		if(length(valids)<=1) continue; end
		R2s(m) = 1.0 - sum((mpredictions(valids)-mdata(valids,2)).^2)./(length(valids)*var(mdata(valids,2)));
	end
end


% return the mean of the flattened array A, only including entries above a threshold
function M = mean_above(A,threshold,default)
	A = reshape(A,prod(size(A)),1);
	A = A(A>threshold);
	if(length(A)==0)
		M = default;
	else
		M = mean(A);
	end
end



function s = sprint_numeric_vector(values,sep)
	s = strjoin(cellfun(@num2str,num2cell(values),'uniformoutput',false),sep);
end


function same = compare_arrays(X,Y)
	if(length(size(X))~=length(size(Y)))
		same = false;
	else
		if(((size(X,1)==1) & (size(Y,2)==1)) | ((size(X,2)==1) & (size(Y,1)==1)))
			% both are 1D arrays, but one is a row-vector and the other is a column-vector, so just rotate one
			Y = Y';
		end
		if(any(size(X)~=size(Y)))
			same = false;
		elseif(any(X~=Y))
			same = false;
		else
			same = true;
		end
	end
end



% for each of N variables, with medians[1:N], mins[1:N], maxs[1:N], choose a random values within [min:max] and with the given median
% Conditional upon drawing a value below the median (which occurs 50% of the time), the value is chosen randomly within [min:median]
% Conditional upon drawing a value above the median (which occurs 50% of the time), the value is chosen randomly within [median:max]
function random_values = randomize_symmetric(medians, mins, maxs)
	N		= size(medians);
	mins	= min(mins,maxs); % make sure min<=max
	medians = max(mins,min(maxs,medians)); % make sure median is between min & max
	
	% for each variable, choose randomly whether to pick a value above or below the median
	aboves = (rand(N)>0.5);
	aboves(mins>=medians) = true; % for variables were [min:median] is trivial, always pick one above the median
	aboves(maxs<=medians) = false; % for variables were [median:max] is trivial, always pick one below the median
	
	% specify random interval based on whether to draw above or below the median
	rmins 			= medians;
	rmaxs 			= medians;
	rmaxs(aboves) 	= maxs(aboves); % for those variables landing above the medians, correct the maximum to maxs
	rmins(~aboves) 	= mins(~aboves); % for those variables landing below the medians, correct the minimum to mins
	
	% pick random values within each variable's random interval
	random_values 	= rmins + rand(N).*(rmaxs-rmins);
end

